@@ -469,10 +469,10 @@ impl BufferContent {
                 }
                 Ok(_) => {
                     if line.text.ends_with('\n') {
-                        line.text.truncate(line.text.len() - 1);
+                        line.text.pop();
                     }
                     if line.text.ends_with('\r') {
-                        line.text.truncate(line.text.len() - 1);
+                        line.text.pop();
                     }
 
                     self.lines.push(line);
@@ -503,7 +503,7 @@ impl BufferContent {
         W: io::Write,
     {
         for line in &self.lines {
-            writeln!(write, "{}", line.as_str())?;
+            write!(write, "{}\n", line.as_str())?;
         }
         Ok(())
     }
@@ -541,7 +541,7 @@ impl BufferContent {
     }
 
     pub fn insert_text(&mut self, position: BufferPosition, text: &str) -> BufferRange {
-        if !text.contains('\n') {
+        if !text.contains(&['\n', '\r'][..]) {
             let line = &mut self.lines[position.line_index as usize];
             let previous_len = line.as_str().len();
             line.insert_text(position.column_byte_index as _, text);
@@ -1293,6 +1293,7 @@ pub struct BufferHandle(pub u32);
 
 pub struct InsertProcess {
     pub alive: bool,
+    pub handle: Option<PlatformProcessHandle>,
     pub buffer_handle: BufferHandle,
     pub position: BufferPosition,
     pub input: Option<PooledBuf>,
@@ -1370,11 +1371,28 @@ impl BufferCollection {
         }
     }
 
-    pub(crate) fn remove_now(&mut self, handle: BufferHandle, word_database: &mut WordDatabase) {
+    pub(crate) fn remove_now(
+        &mut self,
+        platform: &mut Platform,
+        handle: BufferHandle,
+        word_database: &mut WordDatabase,
+    ) {
         let buffer = &mut self.buffers[handle.0 as usize];
         if buffer.alive {
             buffer.dispose(word_database);
         }
+
+        for process in &mut self.insert_processes {
+            if process.buffer_handle != handle {
+                continue;
+            }
+
+            if let Some(handle) = process.handle.take() {
+                platform
+                    .requests
+                    .enqueue(PlatformRequest::KillProcess { handle });
+            }
+        }
     }
 
     pub fn spawn_insert_process(
@@ -1398,6 +1416,7 @@ impl BufferCollection {
                 let index = self.insert_processes.len();
                 self.insert_processes.push(InsertProcess {
                     alive: false,
+                    handle: None,
                     buffer_handle,
                     position,
                     input: None,
@@ -1409,6 +1428,7 @@ impl BufferCollection {
 
         let process = &mut self.insert_processes[index];
         process.alive = true;
+        process.handle = None;
         process.buffer_handle = buffer_handle;
         process.position = position;
         process.input = input;
@@ -1436,7 +1456,10 @@ impl BufferCollection {
         index: u32,
         handle: PlatformProcessHandle,
     ) {
-        if let Some(buf) = self.insert_processes[index as usize].input.take() {
+        let process = &mut self.insert_processes[index as usize];
+        process.handle = Some(handle);
+
+        if let Some(buf) = process.input.take() {
             platform
                 .requests
                 .enqueue(PlatformRequest::WriteToProcess { handle, buf });
@@ -1454,22 +1477,23 @@ impl BufferCollection {
         events: &mut EditorEventQueue,
     ) {
         let process = &mut self.insert_processes[index as usize];
+        if process.handle.is_none() {
+            return;
+        }
 
         let mut buf = Default::default();
         let texts = process.output_residual_bytes.receive_bytes(&mut buf, bytes);
 
         let buffer = &mut self.buffers[process.buffer_handle.0 as usize];
-        if buffer.alive {
-            process.position = buffer.content().saturate_position(process.position);
-            let mut position = process.position;
+        process.position = buffer.content().saturate_position(process.position);
+        let mut position = process.position;
 
-            for text in texts {
-                let insert_range = buffer.insert_text(word_database, position, text, events);
-                position = position.insert(insert_range);
-                for process in &mut self.insert_processes {
-                    if process.buffer_handle == buffer.handle() {
-                        process.position = process.position.insert(insert_range);
-                    }
+        for text in texts {
+            let insert_range = buffer.insert_text(word_database, position, text, events);
+            position = position.insert(insert_range);
+            for process in &mut self.insert_processes {
+                if process.buffer_handle == buffer.handle() {
+                    process.position = process.position.insert(insert_range);
                 }
             }
         }
@@ -1481,27 +1505,10 @@ impl BufferCollection {
         index: u32,
         events: &mut EditorEventQueue,
     ) {
+        self.on_process_output(word_database, index, &[], events);
         let process = &mut self.insert_processes[index as usize];
         process.alive = false;
-
-        let mut buf = Default::default();
-        let texts = process.output_residual_bytes.receive_bytes(&mut buf, &[]);
-
-        let buffer = &mut self.buffers[process.buffer_handle.0 as usize];
-        if buffer.alive {
-            process.position = buffer.content().saturate_position(process.position);
-            let mut position = process.position;
-
-            for text in texts {
-                let insert_range = buffer.insert_text(word_database, position, text, events);
-                position = position.insert(insert_range);
-                for process in &mut self.insert_processes {
-                    if process.buffer_handle == buffer.handle() {
-                        process.position = process.position.insert(insert_range);
-                    }
-                }
-            }
-        }
+        process.handle = None;
     }
 }
 
