@@ -1,7 +1,8 @@
-use std::{fmt, num::NonZeroU8, str::FromStr};
+use std::num::NonZeroU8;
 
 use crate::{
     buffer::{Buffer, BufferCollection, BufferHandle, CharDisplayDistances},
+    buffer_history::EditKind,
     buffer_position::{BufferPosition, BufferPositionIndex, BufferRange},
     client::ClientHandle,
     cursor::{Cursor, CursorCollection},
@@ -30,18 +31,19 @@ pub enum CursorMovementKind {
 }
 
 pub struct BufferView {
+    alive: bool,
+    handle: BufferViewHandle,
     pub client_handle: ClientHandle,
     pub buffer_handle: BufferHandle,
     pub cursors: CursorCollection,
 }
 
 impl BufferView {
-    pub fn new(client_handle: ClientHandle, buffer_handle: BufferHandle) -> Self {
-        Self {
-            client_handle,
-            buffer_handle,
-            cursors: CursorCollection::new(),
-        }
+    fn reset(&mut self, client_handle: ClientHandle, buffer_handle: BufferHandle) {
+        self.alive = true;
+        self.client_handle = client_handle;
+        self.buffer_handle = buffer_handle;
+        self.cursors.mut_guard().clear();
     }
 
     pub fn move_cursors(
@@ -64,10 +66,7 @@ impl BufferView {
             Err(n)
         }
 
-        let buffer = match buffers.get(self.buffer_handle) {
-            Some(buffer) => buffer.content(),
-            None => return,
-        };
+        let buffer = buffers.get(self.buffer_handle).content();
 
         let mut cursors = self.cursors.mut_guard();
         match movement {
@@ -180,8 +179,7 @@ impl BufferView {
                     if let Some(distance) = saved_display_distance {
                         let line = buffer.line_at(c.position.line_index as _).as_str();
                         c.position.column_byte_index = CharDisplayDistances::new(line, tab_size)
-                            .skip_while(|d| d.distance <= distance as _)
-                            .next()
+                            .find(|d| d.distance > distance as _)
                             .map(|d| d.char_index)
                             .unwrap_or(line.len())
                             as _;
@@ -198,8 +196,7 @@ impl BufferView {
                     if let Some(distance) = saved_display_distance {
                         let line = buffer.line_at(c.position.line_index as _).as_str();
                         c.position.column_byte_index = CharDisplayDistances::new(line, tab_size)
-                            .skip_while(|d| d.distance <= distance as _)
-                            .next()
+                            .find(|d| d.distance > distance as _)
                             .map(|d| d.char_index)
                             .unwrap_or(line.len())
                             as _;
@@ -333,32 +330,38 @@ impl BufferView {
         &self,
         buffers: &BufferCollection,
         text: &mut String,
-        ranges: &mut Vec<(u32, u32)>,
-    ) {
-        ranges.clear();
-
-        let buffer = match buffers.get(self.buffer_handle) {
-            Some(buffer) => buffer.content(),
-            None => return,
-        };
+        ranges: &mut [(BufferPositionIndex, BufferPositionIndex)],
+    ) -> usize {
+        let mut ranges_index = 0;
 
+        let buffer = buffers.get(self.buffer_handle).content();
         let mut iter = self.cursors[..].iter();
         if let Some(cursor) = iter.next() {
             let mut last_range = cursor.to_range();
             let from = text.len() as _;
-            buffer.append_range_text_to_string(last_range, text);
-            ranges.push((from, text.len() as _));
+            for t in buffer.text_range(last_range) {
+                text.push_str(t);
+            }
+            ranges[ranges_index] = (from, text.len() as _);
+            ranges_index += 1;
+
             for cursor in iter {
                 let range = cursor.to_range();
                 if range.from.line_index > last_range.to.line_index {
                     text.push('\n');
                 }
                 let from = text.len() as _;
-                buffer.append_range_text_to_string(range, text);
-                ranges.push((from, text.len() as _));
+                for t in buffer.text_range(range) {
+                    text.push_str(t);
+                }
+                ranges[ranges_index] = (from, text.len() as _);
+                ranges_index += 1;
+
                 last_range = range;
             }
         }
+
+        ranges_index
     }
 
     pub fn insert_text_at_cursor_positions(
@@ -368,10 +371,9 @@ impl BufferView {
         text: &str,
         events: &mut EditorEventQueue,
     ) {
-        if let Some(buffer) = buffers.get_mut(self.buffer_handle) {
-            for cursor in self.cursors[..].iter().rev() {
-                buffer.insert_text(word_database, cursor.position, text, events);
-            }
+        let buffer = buffers.get_mut(self.buffer_handle);
+        for cursor in self.cursors[..].iter().rev() {
+            buffer.insert_text(word_database, cursor.position, text, events);
         }
     }
 
@@ -381,10 +383,9 @@ impl BufferView {
         word_database: &mut WordDatabase,
         events: &mut EditorEventQueue,
     ) {
-        if let Some(buffer) = buffers.get_mut(self.buffer_handle) {
-            for cursor in self.cursors[..].iter().rev() {
-                buffer.delete_range(word_database, cursor.to_range(), events);
-            }
+        let buffer = buffers.get_mut(self.buffer_handle);
+        for cursor in self.cursors[..].iter().rev() {
+            buffer.delete_range(word_database, cursor.to_range(), events);
         }
     }
 
@@ -394,11 +395,8 @@ impl BufferView {
         positions: &mut Vec<BufferPosition>,
     ) {
         positions.clear();
-        let buffer = match buffers.get_mut(self.buffer_handle) {
-            Some(buffer) => buffer.content(),
-            None => return,
-        };
 
+        let buffer = buffers.get_mut(self.buffer_handle).content();
         for cursor in self.cursors[..].iter() {
             let position = buffer.position_before(cursor.position);
             let word = buffer.word_at(position);
@@ -417,11 +415,7 @@ impl BufferView {
         positions: &[BufferPosition],
         events: &mut EditorEventQueue,
     ) {
-        let buffer = match buffers.get_mut(self.buffer_handle) {
-            Some(buffer) => buffer,
-            None => return,
-        };
-
+        let buffer = buffers.get_mut(self.buffer_handle);
         for (cursor, &position) in self.cursors[..].iter().zip(positions.iter()).rev() {
             let range = BufferRange::between(position, cursor.position);
             buffer.delete_range(word_database, range, events);
@@ -435,19 +429,28 @@ impl BufferView {
         word_database: &mut WordDatabase,
         events: &mut EditorEventQueue,
     ) {
-        let edits = match buffers.get_mut(self.buffer_handle) {
-            Some(buffer) => buffer.undo(word_database, events),
-            None => return,
-        };
-
-        if let Some(edit) = edits.last() {
-            let mut cursors = self.cursors.mut_guard();
-            cursors.clear();
+        let edits = buffers
+            .get_mut(self.buffer_handle)
+            .undo(word_database, events);
+        let mut cursors = self.cursors.mut_guard();
+        let mut last_edit_kind = None;
+        for edit in edits {
+            if last_edit_kind != Some(edit.kind) {
+                cursors.clear();
+            }
+            let position = match edit.kind {
+                EditKind::Insert => edit.range.to,
+                EditKind::Delete => edit.range.from,
+            };
             cursors.add(Cursor {
                 anchor: edit.range.from,
-                position: edit.range.from,
-            })
+                position,
+            });
+            last_edit_kind = Some(edit.kind);
         }
+
+        events.enqueue_fix_cursors(self.handle, &cursors[..]);
+        cursors.clear();
     }
 
     pub fn redo(
@@ -456,108 +459,86 @@ impl BufferView {
         word_database: &mut WordDatabase,
         events: &mut EditorEventQueue,
     ) {
-        let edits = match buffers.get_mut(self.buffer_handle) {
-            Some(buffer) => buffer.redo(word_database, events),
-            None => return,
-        };
-
-        if let Some(edit) = edits.last() {
-            let mut cursors = self.cursors.mut_guard();
-            cursors.clear();
+        let edits = buffers
+            .get_mut(self.buffer_handle)
+            .redo(word_database, events);
+        let mut cursors = self.cursors.mut_guard();
+        let mut last_edit_kind = None;
+        for edit in edits {
+            if last_edit_kind != Some(edit.kind) {
+                cursors.clear();
+            }
+            let position = match edit.kind {
+                EditKind::Insert => {
+                    for cursor in &mut cursors[..] {
+                        cursor.insert(edit.range);
+                    }
+                    edit.range.to
+                }
+                EditKind::Delete => {
+                    for cursor in &mut cursors[..] {
+                        cursor.delete(edit.range);
+                    }
+                    edit.range.from
+                }
+            };
             cursors.add(Cursor {
                 anchor: edit.range.from,
-                position: edit.range.from,
+                position,
             });
+            last_edit_kind = Some(edit.kind);
         }
+
+        events.enqueue_fix_cursors(self.handle, &cursors[..]);
+        cursors.clear();
     }
 }
 
 #[derive(Clone, Copy, Eq, PartialEq)]
 pub struct BufferViewHandle(u32);
-impl fmt::Display for BufferViewHandle {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        self.0.fmt(f)
-    }
-}
-impl FromStr for BufferViewHandle {
-    type Err = ();
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        match s.parse() {
-            Ok(i) => Ok(Self(i)),
-            Err(_) => Err(()),
-        }
-    }
-}
 
 #[derive(Default)]
 pub struct BufferViewCollection {
-    buffer_views: Vec<Option<BufferView>>,
+    buffer_views: Vec<BufferView>,
 }
 
 impl BufferViewCollection {
-    pub fn add(&mut self, buffer_view: BufferView) -> BufferViewHandle {
-        for (i, slot) in self.buffer_views.iter_mut().enumerate() {
-            if slot.is_none() {
-                *slot = Some(buffer_view);
+    pub fn add_new(
+        &mut self,
+        client_handle: ClientHandle,
+        buffer_handle: BufferHandle,
+    ) -> BufferViewHandle {
+        for (i, view) in self.buffer_views.iter_mut().enumerate() {
+            if !view.alive {
+                view.reset(client_handle, buffer_handle);
                 return BufferViewHandle(i as _);
             }
         }
-
         let handle = BufferViewHandle(self.buffer_views.len() as _);
-        self.buffer_views.push(Some(buffer_view));
+        self.buffer_views.push(BufferView {
+            alive: true,
+            handle,
+            client_handle,
+            buffer_handle,
+            cursors: CursorCollection::new(),
+        });
         handle
     }
 
     pub fn remove_buffer_views(&mut self, buffer_handle: BufferHandle) {
-        for maybe_view in &mut self.buffer_views {
-            if let Some(view) = maybe_view {
-                if view.buffer_handle == buffer_handle {
-                    *maybe_view = None;
-                }
+        for view in &mut self.buffer_views {
+            if view.alive && view.buffer_handle == buffer_handle {
+                view.alive = false;
             }
         }
     }
 
-    pub fn get(&self, handle: BufferViewHandle) -> Option<&BufferView> {
-        self.buffer_views[handle.0 as usize].as_ref()
+    pub fn get(&self, handle: BufferViewHandle) -> &BufferView {
+        &self.buffer_views[handle.0 as usize]
     }
 
-    pub fn get_mut(&mut self, handle: BufferViewHandle) -> Option<&mut BufferView> {
-        self.buffer_views[handle.0 as usize].as_mut()
-    }
-
-    pub fn on_buffer_load(&mut self, buffer: &Buffer) {
-        let buffer_handle = buffer.handle();
-        let buffer = buffer.content();
-
-        for view in self.buffer_views.iter_mut().flatten() {
-            if view.buffer_handle == buffer_handle {
-                for c in &mut view.cursors.mut_guard()[..] {
-                    c.anchor = buffer.saturate_position(c.anchor);
-                    c.position = buffer.saturate_position(c.position);
-                }
-            }
-        }
-    }
-
-    pub fn on_buffer_insert_text(&mut self, buffer_handle: BufferHandle, range: BufferRange) {
-        for view in self.buffer_views.iter_mut().flatten() {
-            if view.buffer_handle == buffer_handle {
-                for c in &mut view.cursors.mut_guard()[..] {
-                    c.insert(range);
-                }
-            }
-        }
-    }
-
-    pub fn on_buffer_delete_text(&mut self, buffer_handle: BufferHandle, range: BufferRange) {
-        for view in self.buffer_views.iter_mut().flatten() {
-            if view.buffer_handle == buffer_handle {
-                for c in &mut view.cursors.mut_guard()[..] {
-                    c.delete(range);
-                }
-            }
-        }
+    pub fn get_mut(&mut self, handle: BufferViewHandle) -> &mut BufferView {
+        &mut self.buffer_views[handle.0 as usize]
     }
 
     pub fn buffer_view_handle_from_buffer_handle(
@@ -569,15 +550,55 @@ impl BufferViewCollection {
             .buffer_views
             .iter()
             .position(|v| {
-                v.as_ref()
-                    .map(|v| v.buffer_handle == buffer_handle && v.client_handle == client_handle)
-                    .unwrap_or(false)
+                v.alive && v.buffer_handle == buffer_handle && v.client_handle == client_handle
             })
             .map(|i| BufferViewHandle(i as _));
 
         match current_buffer_view_handle {
             Some(handle) => handle,
-            None => self.add(BufferView::new(client_handle, buffer_handle)),
+            None => self.add_new(client_handle, buffer_handle),
+        }
+    }
+
+    pub(crate) fn on_buffer_load(&mut self, buffer: &Buffer) {
+        let buffer_handle = buffer.handle();
+        let buffer = buffer.content();
+
+        for view in self.buffer_views.iter_mut().filter(|v| v.alive) {
+            if view.buffer_handle == buffer_handle {
+                for c in &mut view.cursors.mut_guard()[..] {
+                    c.anchor = buffer.saturate_position(c.anchor);
+                    c.position = buffer.saturate_position(c.position);
+                }
+            }
+        }
+    }
+
+    pub(crate) fn on_buffer_insert_text(
+        &mut self,
+        buffer_handle: BufferHandle,
+        range: BufferRange,
+    ) {
+        for view in self.buffer_views.iter_mut().filter(|v| v.alive) {
+            if view.buffer_handle == buffer_handle {
+                for c in &mut view.cursors.mut_guard()[..] {
+                    c.insert(range);
+                }
+            }
+        }
+    }
+
+    pub(crate) fn on_buffer_delete_text(
+        &mut self,
+        buffer_handle: BufferHandle,
+        range: BufferRange,
+    ) {
+        for view in self.buffer_views.iter_mut().filter(|v| v.alive) {
+            if view.buffer_handle == buffer_handle {
+                for c in &mut view.cursors.mut_guard()[..] {
+                    c.delete(range);
+                }
+            }
         }
     }
 }
@@ -586,11 +607,11 @@ impl BufferViewCollection {
 mod tests {
     use super::*;
 
-    use crate::{buffer::BufferCapabilities, buffer_position::BufferPosition};
+    use std::ops::Range;
+
+    use crate::{buffer::BufferProperties, buffer_position::BufferPosition};
 
     struct TestContext {
-        pub word_database: WordDatabase,
-        pub events: EditorEventQueue,
         pub buffers: BufferCollection,
         pub buffer_views: BufferViewCollection,
         pub buffer_view_handle: BufferViewHandle,
@@ -603,7 +624,7 @@ mod tests {
 
             let mut buffers = BufferCollection::default();
             let buffer = buffers.add_new();
-            buffer.capabilities = BufferCapabilities::text();
+            buffer.properties = BufferProperties::text();
             buffer.insert_text(
                 &mut word_database,
                 BufferPosition::zero(),
@@ -611,15 +632,11 @@ mod tests {
                 &mut events,
             );
 
-            let buffer_view =
-                BufferView::new(ClientHandle::from_index(0).unwrap(), buffer.handle());
-
             let mut buffer_views = BufferViewCollection::default();
-            let buffer_view_handle = buffer_views.add(buffer_view);
+            let buffer_view_handle =
+                buffer_views.add_new(ClientHandle::from_index(0).unwrap(), buffer.handle());
 
             Self {
-                word_database,
-                events,
                 buffers,
                 buffer_views,
                 buffer_view_handle,
@@ -630,7 +647,7 @@ mod tests {
     #[test]
     fn buffer_view_cursor_movement() {
         fn set_cursor(ctx: &mut TestContext, position: BufferPosition) {
-            let buffer_view = ctx.buffer_views.get_mut(ctx.buffer_view_handle).unwrap();
+            let buffer_view = ctx.buffer_views.get_mut(ctx.buffer_view_handle);
             let mut cursors = buffer_view.cursors.mut_guard();
             cursors.clear();
             cursors.add(Cursor {
@@ -642,7 +659,6 @@ mod tests {
         fn main_cursor_position(ctx: &TestContext) -> BufferPosition {
             ctx.buffer_views
                 .get(ctx.buffer_view_handle)
-                .unwrap()
                 .cursors
                 .main_cursor()
                 .position
@@ -650,14 +666,16 @@ mod tests {
 
         fn assert_movement(
             ctx: &mut TestContext,
-            from: (usize, usize),
-            to: (usize, usize),
+            from: Range<usize>,
+            to: Range<usize>,
             movement: CursorMovement,
         ) {
-            set_cursor(ctx, BufferPosition::line_col(from.0 as _, from.1 as _));
+            set_cursor(
+                ctx,
+                BufferPosition::line_col(from.start as _, from.end as _),
+            );
             ctx.buffer_views
                 .get_mut(ctx.buffer_view_handle)
-                .unwrap()
                 .move_cursors(
                     &ctx.buffers,
                     movement,
@@ -665,61 +683,51 @@ mod tests {
                     NonZeroU8::new(4).unwrap(),
                 );
             assert_eq!(
-                BufferPosition::line_col(to.0 as _, to.1 as _),
+                BufferPosition::line_col(to.start as _, to.end as _),
                 main_cursor_position(ctx)
             );
         }
 
         let mut ctx = TestContext::with_buffer("ab\nc e\nefgh\ni k\nlm");
-        assert_movement(&mut ctx, (2, 2), (2, 2), CursorMovement::ColumnsForward(0));
-        assert_movement(&mut ctx, (2, 2), (2, 3), CursorMovement::ColumnsForward(1));
-        assert_movement(&mut ctx, (2, 2), (2, 4), CursorMovement::ColumnsForward(2));
-        assert_movement(&mut ctx, (2, 2), (3, 0), CursorMovement::ColumnsForward(3));
-        assert_movement(&mut ctx, (2, 2), (3, 3), CursorMovement::ColumnsForward(6));
-        assert_movement(&mut ctx, (2, 2), (4, 0), CursorMovement::ColumnsForward(7));
-        assert_movement(
-            &mut ctx,
-            (2, 2),
-            (4, 2),
-            CursorMovement::ColumnsForward(999),
-        );
+        assert_movement(&mut ctx, 2..2, 2..2, CursorMovement::ColumnsForward(0));
+        assert_movement(&mut ctx, 2..2, 2..3, CursorMovement::ColumnsForward(1));
+        assert_movement(&mut ctx, 2..2, 2..4, CursorMovement::ColumnsForward(2));
+        assert_movement(&mut ctx, 2..2, 3..0, CursorMovement::ColumnsForward(3));
+        assert_movement(&mut ctx, 2..2, 3..3, CursorMovement::ColumnsForward(6));
+        assert_movement(&mut ctx, 2..2, 4..0, CursorMovement::ColumnsForward(7));
+        assert_movement(&mut ctx, 2..2, 4..2, CursorMovement::ColumnsForward(999));
 
-        assert_movement(&mut ctx, (2, 2), (2, 2), CursorMovement::ColumnsBackward(0));
-        assert_movement(&mut ctx, (2, 2), (2, 1), CursorMovement::ColumnsBackward(1));
-        assert_movement(&mut ctx, (2, 0), (1, 3), CursorMovement::ColumnsBackward(1));
-        assert_movement(&mut ctx, (2, 2), (1, 3), CursorMovement::ColumnsBackward(3));
-        assert_movement(&mut ctx, (2, 2), (0, 2), CursorMovement::ColumnsBackward(7));
-        assert_movement(
-            &mut ctx,
-            (2, 2),
-            (0, 0),
-            CursorMovement::ColumnsBackward(999),
-        );
+        assert_movement(&mut ctx, 2..2, 2..2, CursorMovement::ColumnsBackward(0));
+        assert_movement(&mut ctx, 2..2, 2..1, CursorMovement::ColumnsBackward(1));
+        assert_movement(&mut ctx, 2..0, 1..3, CursorMovement::ColumnsBackward(1));
+        assert_movement(&mut ctx, 2..2, 1..3, CursorMovement::ColumnsBackward(3));
+        assert_movement(&mut ctx, 2..2, 0..2, CursorMovement::ColumnsBackward(7));
+        assert_movement(&mut ctx, 2..2, 0..0, CursorMovement::ColumnsBackward(999));
 
-        assert_movement(&mut ctx, (2, 2), (2, 2), CursorMovement::WordsForward(0));
-        assert_movement(&mut ctx, (2, 0), (2, 4), CursorMovement::WordsForward(1));
-        assert_movement(&mut ctx, (2, 0), (3, 0), CursorMovement::WordsForward(2));
-        assert_movement(&mut ctx, (2, 2), (3, 2), CursorMovement::WordsForward(3));
-        assert_movement(&mut ctx, (2, 2), (3, 3), CursorMovement::WordsForward(4));
-        assert_movement(&mut ctx, (2, 2), (4, 0), CursorMovement::WordsForward(5));
-        assert_movement(&mut ctx, (2, 2), (4, 2), CursorMovement::WordsForward(6));
-        assert_movement(&mut ctx, (2, 2), (4, 2), CursorMovement::WordsForward(999));
+        assert_movement(&mut ctx, 2..2, 2..2, CursorMovement::WordsForward(0));
+        assert_movement(&mut ctx, 2..0, 2..4, CursorMovement::WordsForward(1));
+        assert_movement(&mut ctx, 2..0, 3..0, CursorMovement::WordsForward(2));
+        assert_movement(&mut ctx, 2..2, 3..2, CursorMovement::WordsForward(3));
+        assert_movement(&mut ctx, 2..2, 3..3, CursorMovement::WordsForward(4));
+        assert_movement(&mut ctx, 2..2, 4..0, CursorMovement::WordsForward(5));
+        assert_movement(&mut ctx, 2..2, 4..2, CursorMovement::WordsForward(6));
+        assert_movement(&mut ctx, 2..2, 4..2, CursorMovement::WordsForward(999));
 
-        assert_movement(&mut ctx, (2, 2), (2, 2), CursorMovement::WordsBackward(0));
-        assert_movement(&mut ctx, (2, 0), (1, 3), CursorMovement::WordsBackward(1));
-        assert_movement(&mut ctx, (2, 0), (1, 2), CursorMovement::WordsBackward(2));
-        assert_movement(&mut ctx, (2, 2), (2, 0), CursorMovement::WordsBackward(1));
-        assert_movement(&mut ctx, (2, 2), (1, 3), CursorMovement::WordsBackward(2));
-        assert_movement(&mut ctx, (2, 2), (1, 2), CursorMovement::WordsBackward(3));
-        assert_movement(&mut ctx, (2, 2), (1, 0), CursorMovement::WordsBackward(4));
-        assert_movement(&mut ctx, (2, 2), (0, 2), CursorMovement::WordsBackward(5));
-        assert_movement(&mut ctx, (2, 2), (0, 0), CursorMovement::WordsBackward(6));
-        assert_movement(&mut ctx, (2, 2), (0, 0), CursorMovement::WordsBackward(999));
+        assert_movement(&mut ctx, 2..2, 2..2, CursorMovement::WordsBackward(0));
+        assert_movement(&mut ctx, 2..0, 1..3, CursorMovement::WordsBackward(1));
+        assert_movement(&mut ctx, 2..0, 1..2, CursorMovement::WordsBackward(2));
+        assert_movement(&mut ctx, 2..2, 2..0, CursorMovement::WordsBackward(1));
+        assert_movement(&mut ctx, 2..2, 1..3, CursorMovement::WordsBackward(2));
+        assert_movement(&mut ctx, 2..2, 1..2, CursorMovement::WordsBackward(3));
+        assert_movement(&mut ctx, 2..2, 1..0, CursorMovement::WordsBackward(4));
+        assert_movement(&mut ctx, 2..2, 0..2, CursorMovement::WordsBackward(5));
+        assert_movement(&mut ctx, 2..2, 0..0, CursorMovement::WordsBackward(6));
+        assert_movement(&mut ctx, 2..2, 0..0, CursorMovement::WordsBackward(999));
 
         let mut ctx = TestContext::with_buffer("123\n  abc def\nghi");
-        assert_movement(&mut ctx, (1, 0), (1, 2), CursorMovement::WordsForward(1));
-        assert_movement(&mut ctx, (1, 9), (2, 0), CursorMovement::WordsForward(1));
-        assert_movement(&mut ctx, (1, 2), (1, 0), CursorMovement::WordsBackward(1));
-        assert_movement(&mut ctx, (2, 0), (1, 9), CursorMovement::WordsBackward(1));
+        assert_movement(&mut ctx, 1..0, 1..2, CursorMovement::WordsForward(1));
+        assert_movement(&mut ctx, 1..9, 2..0, CursorMovement::WordsForward(1));
+        assert_movement(&mut ctx, 1..2, 1..0, CursorMovement::WordsBackward(1));
+        assert_movement(&mut ctx, 2..0, 1..9, CursorMovement::WordsBackward(1));
     }
 }
