@@ -4,29 +4,29 @@ use std::{
         io::{AsRawFd, RawFd},
         net::{UnixListener, UnixStream},
     },
-    sync::atomic::{AtomicIsize, Ordering},
     time::Duration,
 };
 
-use pepper::{
-    application::{AnyError, ApplicationEvent, ClientApplication, ServerApplication},
+use crate::{
+    application::{ApplicationConfig, ClientApplication, ServerApplication},
     client::ClientHandle,
-    platform::{BufPool, Key, Platform, PlatformRequest, ProcessHandle},
+    platform::{Key, PlatformEvent, PlatformProcessHandle, PlatformRequest},
     Args,
 };
 
 mod unix_utils;
 use unix_utils::{
-    get_terminal_size, is_pipped, parse_terminal_keys, read, read_from_connection, run,
-    suspend_process, Process, RawMode,
+    is_pipped, read, read_from_connection, run, suspend_process, write_all_bytes, Process, Terminal,
 };
 
 const MAX_CLIENT_COUNT: usize = 20;
-const MAX_PROCESS_COUNT: usize = 42;
+const MAX_PROCESS_COUNT: usize = 43;
 const MAX_TRIGGERED_EVENT_COUNT: usize = 32;
 
-pub fn main() {
-    run(run_server, run_client);
+pub fn try_launching_debugger() {}
+
+pub fn main(config: ApplicationConfig) {
+    run(config, run_server, run_client);
 }
 
 fn errno() -> libc::c_int {
@@ -35,7 +35,6 @@ fn errno() -> libc::c_int {
 
 enum Event {
     Resize,
-    FlushRequests(bool),
     Fd(RawFd),
 }
 impl Event {
@@ -49,14 +48,6 @@ impl Event {
                 data: 0,
                 udata: index as _,
             },
-            Self::FlushRequests(triggered) => libc::kevent {
-                ident: 0,
-                filter: libc::EVFILT_USER,
-                flags: flags | libc::EV_ONESHOT,
-                fflags: if triggered { libc::NOTE_TRIGGER } else { 0 },
-                data: 0,
-                udata: index as _,
-            },
             Self::Fd(fd) => libc::kevent {
                 ident: fd as _,
                 filter: libc::EVFILT_READ,
@@ -176,144 +167,52 @@ impl Drop for Kqueue {
     }
 }
 
-fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
+fn run_server(config: ApplicationConfig, listener: UnixListener) {
     use io::Write;
 
     const NONE_PROCESS: Option<Process> = None;
-    static KQUEUE_FD: AtomicIsize = AtomicIsize::new(-1);
 
-    let kqueue = Kqueue::new();
-    kqueue.add(Event::FlushRequests(false), 0);
-    kqueue.add(Event::Fd(listener.as_raw_fd()), 1);
-    let mut kqueue_events = KqueueEvents::new();
-
-    KQUEUE_FD.store(kqueue.as_raw_fd() as _, Ordering::Relaxed);
-
-    fn flush_requests() {
-        let fd = KQUEUE_FD.load(Ordering::Relaxed) as _;
-        let event = Event::FlushRequests(true).into_kevent(libc::EV_ADD, 0);
-        if !modify_kqueue(fd, &event) {
-            panic!("error trigerring flush events");
-        }
-    }
-
-    let (request_sender, request_receiver) = ServerApplication::platform_request_channel();
-    let platform = Platform::new(flush_requests, request_sender);
-    let event_sender = ServerApplication::run(args, platform);
+    let mut application = match ServerApplication::new(config) {
+        Some(application) => application,
+        None => return,
+    };
 
     let mut client_connections: [Option<UnixStream>; MAX_CLIENT_COUNT] = Default::default();
     let mut processes = [NONE_PROCESS; MAX_PROCESS_COUNT];
-    let mut buf_pool = BufPool::default();
 
-    let mut timeout = Some(ServerApplication::idle_duration());
+    let mut events = Vec::new();
+    let mut timeout = None;
 
-    const CLIENTS_START_INDEX: usize = 1 + 1;
+    const CLIENTS_START_INDEX: usize = 1;
     const CLIENTS_LAST_INDEX: usize = CLIENTS_START_INDEX + MAX_CLIENT_COUNT - 1;
     const PROCESSES_START_INDEX: usize = CLIENTS_LAST_INDEX + 1;
     const PROCESSES_LAST_INDEX: usize = PROCESSES_START_INDEX + MAX_PROCESS_COUNT - 1;
 
+    let kqueue = Kqueue::new();
+    kqueue.add(Event::Fd(listener.as_raw_fd()), 0);
+    let mut kqueue_events = KqueueEvents::new();
+
     loop {
-        let events = kqueue.wait(&mut kqueue_events, timeout);
-        if events.len() == 0 {
-            timeout = None;
-            event_sender.send(ApplicationEvent::Idle)?;
-            continue;
+        let kqueue_events = kqueue.wait(&mut kqueue_events, timeout);
+        if kqueue_events.len() == 0 {
+            match timeout {
+                Some(Duration::ZERO) => timeout = Some(ServerApplication::idle_duration()),
+                Some(_) => {
+                    events.push(PlatformEvent::Idle);
+                    timeout = None;
+                }
+                None => unreachable!(),
+            }
         }
 
-        for event in events {
+        for event in kqueue_events {
             let (event_index, event_data) = match event {
                 Ok(event) => (event.index, event.data),
-                Err(()) => return Ok(()),
+                Err(()) => return,
             };
+
             match event_index {
                 0 => {
-                    kqueue.add(Event::FlushRequests(false), 0);
-                    for request in request_receiver.try_iter() {
-                        match request {
-                            PlatformRequest::Exit => return Ok(()),
-                            PlatformRequest::WriteToClient { handle, buf } => {
-                                let index = handle.into_index();
-                                if let Some(ref mut connection) = client_connections[index] {
-                                    if connection.write_all(buf.as_bytes()).is_err() {
-                                        kqueue.remove(Event::Fd(connection.as_raw_fd()));
-                                        client_connections[index] = None;
-                                        event_sender
-                                            .send(ApplicationEvent::ConnectionClose { handle })?;
-                                    }
-                                }
-                            }
-                            PlatformRequest::CloseClient { handle } => {
-                                let index = handle.into_index();
-                                if let Some(connection) = client_connections[index].take() {
-                                    kqueue.remove(Event::Fd(connection.as_raw_fd()));
-                                }
-                                event_sender.send(ApplicationEvent::ConnectionClose { handle })?;
-                            }
-                            PlatformRequest::SpawnProcess {
-                                tag,
-                                mut command,
-                                buf_len,
-                            } => {
-                                let mut spawned = false;
-                                for (i, p) in processes.iter_mut().enumerate() {
-                                    if p.is_some() {
-                                        continue;
-                                    }
-
-                                    let handle = ProcessHandle(i);
-                                    if let Ok(child) = command.spawn() {
-                                        let process = Process::new(child, tag, buf_len);
-                                        if let Some(fd) = process.try_as_raw_fd() {
-                                            kqueue.add(Event::Fd(fd), PROCESSES_START_INDEX + i);
-                                        }
-                                        *p = Some(process);
-                                        event_sender.send(ApplicationEvent::ProcessSpawned {
-                                            tag,
-                                            handle,
-                                        })?;
-                                        spawned = true;
-                                    }
-                                    break;
-                                }
-                                if !spawned {
-                                    event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                                }
-                            }
-                            PlatformRequest::WriteToProcess { handle, buf } => {
-                                let index = handle.0;
-                                if let Some(ref mut process) = processes[index] {
-                                    if !process.write(buf.as_bytes()) {
-                                        if let Some(fd) = process.try_as_raw_fd() {
-                                            kqueue.remove(Event::Fd(fd));
-                                        }
-                                        let tag = process.tag();
-                                        process.kill();
-                                        processes[index] = None;
-                                        event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                                    }
-                                }
-                            }
-                            PlatformRequest::CloseProcessInput { handle } => {
-                                if let Some(ref mut process) = processes[handle.0] {
-                                    process.close_input();
-                                }
-                            }
-                            PlatformRequest::KillProcess { handle } => {
-                                let index = handle.0;
-                                if let Some(ref mut process) = processes[index] {
-                                    if let Some(fd) = process.try_as_raw_fd() {
-                                        kqueue.remove(Event::Fd(fd));
-                                    }
-                                    let tag = process.tag();
-                                    process.kill();
-                                    processes[index] = None;
-                                    event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                                }
-                            }
-                        }
-                    }
-                }
-                1 => {
                     for _ in 0..event_data {
                         match listener.accept() {
                             Ok((connection, _)) => {
@@ -325,8 +224,7 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
                                         );
                                         *c = Some(connection);
                                         let handle = ClientHandle::from_index(i).unwrap();
-                                        event_sender
-                                            .send(ApplicationEvent::ConnectionOpen { handle })?;
+                                        events.push(PlatformEvent::ConnectionOpen { handle });
                                         break;
                                     }
                                 }
@@ -339,43 +237,139 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
                     let index = event_index - CLIENTS_START_INDEX;
                     if let Some(ref mut connection) = client_connections[index] {
                         let handle = ClientHandle::from_index(index).unwrap();
-                        match read_from_connection(connection, &mut buf_pool, event_data as _) {
-                            Ok(buf) if !buf.as_bytes().is_empty() => {
-                                event_sender
-                                    .send(ApplicationEvent::ConnectionOutput { handle, buf })?;
-                            }
-                            _ => {
+                        match read_from_connection(
+                            connection,
+                            &mut application.ctx.platform.buf_pool,
+                            event_data as _,
+                        ) {
+                            Ok(buf) => events.push(PlatformEvent::ConnectionOutput { handle, buf }),
+                            Err(()) => {
                                 kqueue.remove(Event::Fd(connection.as_raw_fd()));
                                 client_connections[index] = None;
-                                event_sender.send(ApplicationEvent::ConnectionClose { handle })?;
+                                events.push(PlatformEvent::ConnectionClose { handle });
                             }
                         }
                     }
-
-                    timeout = Some(ServerApplication::idle_duration());
                 }
                 PROCESSES_START_INDEX..=PROCESSES_LAST_INDEX => {
                     let index = event_index - PROCESSES_START_INDEX;
                     if let Some(ref mut process) = processes[index] {
                         let tag = process.tag();
-                        match process.read(&mut buf_pool) {
+                        match process.read(&mut application.ctx.platform.buf_pool) {
                             Ok(None) => (),
-                            Ok(Some(buf)) if !buf.as_bytes().is_empty() => {
-                                event_sender.send(ApplicationEvent::ProcessExit { tag })?;
-                            }
-                            _ => {
+                            Ok(Some(buf)) => events.push(PlatformEvent::ProcessExit { tag }),
+                            Err(()) => {
                                 if let Some(fd) = process.try_as_raw_fd() {
                                     kqueue.remove(Event::Fd(fd));
                                 }
                                 process.kill();
                                 processes[index] = None;
-                                event_sender.send(ApplicationEvent::ProcessExit { tag })?;
+                                events.push(PlatformEvent::ProcessExit { tag });
                             }
                         }
                     }
                 }
                 _ => unreachable!(),
             }
+
+            application.update(events.drain(..));
+            let mut requests = application.ctx.platform.requests.drain();
+            while let Some(request) = requests.next() {
+                match request {
+                    PlatformRequest::Quit => {
+                        for request in requests {
+                            if let PlatformRequest::WriteToClient { buf, .. }
+                            | PlatformRequest::WriteToProcess { buf, .. } = request
+                            {
+                                application.ctx.platform.buf_pool.release(buf);
+                            }
+                        }
+                        return;
+                    }
+                    PlatformRequest::Redraw => timeout = Some(Duration::ZERO),
+                    PlatformRequest::WriteToClient { handle, buf } => {
+                        let index = handle.into_index();
+                        if let Some(ref mut connection) = client_connections[index] {
+                            if connection.write_all(buf.as_bytes()).is_err() {
+                                kqueue.remove(Event::Fd(connection.as_raw_fd()));
+                                client_connections[index] = None;
+                                events.push(PlatformEvent::ConnectionClose { handle });
+                            }
+                        }
+                        application.ctx.platform.buf_pool.release(buf);
+                    }
+                    PlatformRequest::CloseClient { handle } => {
+                        let index = handle.into_index();
+                        if let Some(connection) = client_connections[index].take() {
+                            kqueue.remove(Event::Fd(connection.as_raw_fd()));
+                        }
+                        events.push(PlatformEvent::ConnectionClose { handle });
+                    }
+                    PlatformRequest::SpawnProcess {
+                        tag,
+                        mut command,
+                        buf_len,
+                    } => {
+                        let mut spawned = false;
+                        for (i, p) in processes.iter_mut().enumerate() {
+                            if p.is_some() {
+                                continue;
+                            }
+
+                            let handle = PlatformProcessHandle(i as _);
+                            if let Ok(child) = command.spawn() {
+                                let process = Process::new(child, tag, buf_len);
+                                if let Some(fd) = process.try_as_raw_fd() {
+                                    kqueue.add(Event::Fd(fd), PROCESSES_START_INDEX + i);
+                                }
+                                *p = Some(process);
+                                events.push(PlatformEvent::ProcessSpawned { tag, handle });
+                                spawned = true;
+                            }
+                            break;
+                        }
+                        if !spawned {
+                            events.push(PlatformEvent::ProcessExit { tag });
+                        }
+                    }
+                    PlatformRequest::WriteToProcess { handle, buf } => {
+                        let index = handle.0 as usize;
+                        if let Some(ref mut process) = processes[index] {
+                            if !process.write(buf.as_bytes()) {
+                                if let Some(fd) = process.try_as_raw_fd() {
+                                    kqueue.remove(Event::Fd(fd));
+                                }
+                                let tag = process.tag();
+                                process.kill();
+                                processes[index] = None;
+                                events.push(PlatformEvent::ProcessExit { tag });
+                            }
+                        }
+                        application.ctx.platform.buf_pool.release(buf);
+                    }
+                    PlatformRequest::CloseProcessInput { handle } => {
+                        if let Some(ref mut process) = processes[handle.0 as usize] {
+                            process.close_input();
+                        }
+                    }
+                    PlatformRequest::KillProcess { handle } => {
+                        let index = handle.0 as usize;
+                        if let Some(ref mut process) = processes[index] {
+                            if let Some(fd) = process.try_as_raw_fd() {
+                                kqueue.remove(Event::Fd(fd));
+                            }
+                            let tag = process.tag();
+                            process.kill();
+                            processes[index] = None;
+                            events.push(PlatformEvent::ProcessExit { tag });
+                        }
+                    }
+                }
+            }
+
+            if !events.is_empty() {
+                timeout = Some(Duration::ZERO);
+            }
         }
     }
 }
@@ -383,85 +377,87 @@ fn run_server(args: Args, listener: UnixListener) -> Result<(), AnyError> {
 fn run_client(args: Args, mut connection: UnixStream) {
     use io::{Read, Write};
 
-    let mut buf = [0; 2];
-    match connection.read_exact(&mut buf) {
-        Ok(()) => (),
-        _ => return,
-    }
-    let is_first_client = buf[0] != 0;
-    let client_index = buf[1];
+    let terminal = if args.quit {
+        None
+    } else {
+        Some(Terminal::new())
+    };
 
-    let client_handle = ClientHandle::from_index(client_index as _).unwrap();
-    let is_pipped = is_pipped();
-
-    let stdout = io::stdout();
-    let mut application = ClientApplication::new(client_handle, stdout.lock(), is_pipped);
-    let bytes = application.init(args, is_first_client);
+    let mut application = ClientApplication::new(terminal.as_ref().map(Terminal::to_file));
+    let bytes = application.init(args);
     if connection.write_all(bytes).is_err() {
         return;
     }
 
-    let mut raw_mode;
-
     let kqueue = Kqueue::new();
-    kqueue.add(Event::Fd(connection.as_raw_fd()), 0);
-    kqueue.add(Event::Fd(libc::STDIN_FILENO), 1);
+    kqueue.add(Event::Fd(connection.as_raw_fd()), 1);
+    if is_pipped(libc::STDIN_FILENO) {
+        kqueue.add(Event::Fd(libc::STDIN_FILENO), 3);
+    }
+
     let mut kqueue_events = KqueueEvents::new();
 
-    if is_pipped {
-        raw_mode = None;
-    } else {
-        raw_mode = Some(RawMode::enter());
+    if let Some(terminal) = &terminal {
+        terminal.enter_raw_mode();
+        kqueue.add(Event::Fd(terminal.as_raw_fd()), 0);
+
         kqueue.add(Event::Resize, 2);
 
-        let size = get_terminal_size();
-        let (_, bytes) = application.update(Some(size), &[Key::None], &[], &[]);
+        let size = terminal.get_size();
+        let (_, bytes) = application.update(Some(size), &[Key::None], None, &[]);
+        if connection.write_all(bytes).is_err() {
+            return;
+        }
+    }
+
+    if is_pipped(libc::STDOUT_FILENO) {
+        let (_, bytes) = application.update(None, &[], Some(&[]), &[]);
         if connection.write_all(bytes).is_err() {
             return;
         }
     }
 
-    let backspace_code = match raw_mode {
-        Some(ref raw) => raw.backspace_code(),
-        None => 0,
-    };
     let mut keys = Vec::new();
     let mut buf = Vec::new();
 
     'main_loop: loop {
         for event in kqueue.wait(&mut kqueue_events, None) {
             let mut resize = None;
-            let mut stdin_bytes = &[][..];
+            let mut stdin_bytes = None;
             let mut server_bytes = &[][..];
 
             keys.clear();
 
             match event {
                 Ok(TriggeredEvent { index: 0, data }) => {
+                    if let Some(terminal) = &terminal {
+                        buf.resize(data as _, 0);
+                        match read(terminal.as_raw_fd(), &mut buf) {
+                            Ok(0) | Err(()) => break 'main_loop,
+                            Ok(len) => terminal.parse_keys(&buf[..len], &mut keys),
+                        }
+                    }
+                }
+                Ok(TriggeredEvent { index: 1, data }) => {
                     buf.resize(data as _, 0);
                     match connection.read(&mut buf) {
                         Ok(0) | Err(_) => break 'main_loop,
                         Ok(len) => server_bytes = &buf[..len],
                     }
                 }
-                Ok(TriggeredEvent { index: 1, data }) => {
+                Ok(TriggeredEvent { index: 2, .. }) => {
+                    resize = terminal.as_ref().map(Terminal::get_size);
+                }
+                Ok(TriggeredEvent { index: 3, data }) => {
                     buf.resize(data as _, 0);
                     match read(libc::STDIN_FILENO, &mut buf) {
                         Ok(0) | Err(()) => {
                             kqueue.remove(Event::Fd(libc::STDIN_FILENO));
-                            continue;
-                        }
-                        Ok(len) => {
-                            let bytes = &buf[..len];
-                            if is_pipped {
-                                stdin_bytes = bytes;
-                            } else {
-                                parse_terminal_keys(bytes, backspace_code, &mut keys);
-                            }
+                            stdin_bytes = Some(&[][..]);
                         }
+                        Ok(len) => stdin_bytes = Some(&buf[..len]),
                     }
                 }
-                Ok(TriggeredEvent { index: 2, .. }) => resize = Some(get_terminal_size()),
                 Ok(_) => unreachable!(),
                 Err(()) => break 'main_loop,
             }
@@ -471,11 +467,16 @@ fn run_client(args: Args, mut connection: UnixStream) {
                 break;
             }
             if suspend {
-                suspend_process(&mut application, &mut raw_mode);
+                suspend_process(&mut application, &terminal);
             }
         }
     }
 
-    drop(raw_mode);
-}
+    if is_pipped(libc::STDOUT_FILENO) {
+        let bytes = application.get_stdout_bytes();
+        write_all_bytes(libc::STDOUT_FILENO, bytes);
+    }
 
+    drop(terminal);
+    drop(application);
+}
