@@ -1,3 +1,5 @@
+use std::str::FromStr;
+
 use crate::{
     buffer::BufferContent,
     buffer_position::{BufferPositionIndex, BufferRange},
@@ -19,6 +21,21 @@ pub enum TokenKind {
     Text,
     Whitespace,
 }
+impl FromStr for TokenKind {
+    type Err = ();
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        match s {
+            "keywords" => Ok(Self::Keyword),
+            "types" => Ok(Self::Type),
+            "symbols" => Ok(Self::Symbol),
+            "literals" => Ok(Self::Literal),
+            "strings" => Ok(Self::String),
+            "comments" => Ok(Self::Comment),
+            "texts" => Ok(Self::Text),
+            _ => Err(()),
+        }
+    }
+}
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Token {
@@ -27,7 +44,6 @@ pub struct Token {
     pub to: BufferPositionIndex,
 }
 impl Token {
-    #[inline]
     pub fn contains(&self, column_byte_index: BufferPositionIndex) -> bool {
         self.from <= column_byte_index && column_byte_index < self.to
     }
@@ -80,9 +96,15 @@ impl Syntax {
         }
     }
 
-    pub fn set_glob(&mut self, pattern: &str) -> Result<(), InvalidGlobError> {
-        self.glob_hash = hash_bytes(pattern.as_bytes());
-        self.glob.compile(pattern)
+    fn clear_rules(&mut self) {
+        for r in &mut self.rules {
+            r.clear();
+        }
+    }
+
+    fn set_glob(&mut self, glob: &str, glob_hash: u64) -> Result<(), InvalidGlobError> {
+        self.glob_hash = glob_hash;
+        self.glob.compile(glob)
     }
 
     pub fn set_rule(&mut self, kind: TokenKind, pattern: &str) -> Result<(), PatternError> {
@@ -97,28 +119,27 @@ impl Syntax {
     ) -> LineParseState {
         tokens.clear();
 
-        let line_len = line.len();
         let mut index = 0;
 
         match previous_parse_state {
             LineParseState::Dirty => unreachable!(),
             LineParseState::Finished => (),
             LineParseState::Unfinished(kind, state) => {
-                match self.rules[kind as usize].matches_with_state(line, state) {
-                    MatchResult::Ok(len) => {
+                match self.rules[kind as usize].matches_with_state(line, 0, state) {
+                    MatchResult::Ok(end) => {
                         tokens.push(Token {
                             kind,
                             from: 0,
-                            to: len as _,
+                            to: end as _,
                         });
-                        index += len;
+                        index = end;
                     }
                     MatchResult::Err => (),
                     MatchResult::Pending(state) => {
                         tokens.push(Token {
                             kind,
                             from: 0,
-                            to: line_len as _,
+                            to: line.len() as _,
                         });
                         return LineParseState::Unfinished(kind, state);
                     }
@@ -126,18 +147,17 @@ impl Syntax {
             }
         }
 
-        while index < line_len {
-            let line_slice = &line[index..];
-            let whitespace_len = line_slice
+        while index < line.len() {
+            let from = index;
+            index += line[from..]
                 .bytes()
                 .take_while(u8::is_ascii_whitespace)
                 .count();
-            let line_slice = &line_slice[whitespace_len..];
 
             let mut best_pattern_kind = TokenKind::Text;
-            let mut max_len = 0;
+            let mut max_end = index;
 
-            const ALL_NON_WHITESPACE_TOKEN_KINDS: [TokenKind; 7] = [
+            static ALL_NON_WHITESPACE_TOKEN_KINDS: [TokenKind; 7] = [
                 TokenKind::Keyword,
                 TokenKind::Type,
                 TokenKind::Symbol,
@@ -147,12 +167,12 @@ impl Syntax {
                 TokenKind::Text,
             ];
 
-            for &kind in ALL_NON_WHITESPACE_TOKEN_KINDS.iter() {
+            for kind in ALL_NON_WHITESPACE_TOKEN_KINDS {
                 let pattern = &self.rules[kind as usize];
-                match pattern.matches(line_slice) {
-                    MatchResult::Ok(len) => {
-                        if len > max_len {
-                            max_len = len;
+                match pattern.matches(line, index) {
+                    MatchResult::Ok(end) => {
+                        if end > max_end {
+                            max_end = end;
                             best_pattern_kind = kind;
                         }
                     }
@@ -160,8 +180,8 @@ impl Syntax {
                     MatchResult::Pending(state) => {
                         tokens.push(Token {
                             kind,
-                            from: index as _,
-                            to: line_len as _,
+                            from: from as _,
+                            to: line.len() as _,
                         });
                         return LineParseState::Unfinished(kind, state);
                     }
@@ -170,23 +190,21 @@ impl Syntax {
 
             let mut kind = best_pattern_kind;
 
-            if max_len == 0 {
+            if max_end == index {
                 kind = TokenKind::Text;
-                max_len = line_slice
-                    .bytes()
-                    .take_while(u8::is_ascii_alphanumeric)
+                max_end += line.as_bytes()[index..]
+                    .iter()
+                    .take_while(|b| b.is_ascii_alphanumeric())
                     .count()
                     .max(1);
+
+                max_end = max_end.min(line.len());
+                while !line.is_char_boundary(max_end) {
+                    max_end += 1;
+                }
             }
 
-            max_len += whitespace_len;
-
-            let from = index;
-            index = line_len.min(index + max_len);
-
-            while !line.is_char_boundary(index) {
-                index += 1;
-            }
+            index = max_end;
 
             tokens.push(Token {
                 kind,
@@ -200,17 +218,19 @@ impl Syntax {
 }
 
 #[derive(Clone, Copy, Default, PartialEq, Eq)]
-pub struct SyntaxHandle(usize);
+pub struct SyntaxHandle(u32);
 
 pub struct SyntaxCollection {
     syntaxes: Vec<Syntax>,
+    current_syntax_index: u32,
 }
 
 impl SyntaxCollection {
     pub fn new() -> Self {
-        let mut syntaxes = Vec::new();
-        syntaxes.push(Syntax::new());
-        Self { syntaxes }
+        Self {
+            syntaxes: vec![Syntax::new()],
+            current_syntax_index: 0,
+        }
     }
 
     pub fn find_handle_by_path(&self, path: &str) -> Option<SyntaxHandle> {
@@ -218,25 +238,36 @@ impl SyntaxCollection {
         iter.next();
         for (i, syntax) in iter {
             if syntax.glob.matches(path) {
-                return Some(SyntaxHandle(i));
+                return Some(SyntaxHandle(i as _));
             }
         }
 
         None
     }
 
-    pub fn add(&mut self, syntax: Syntax) {
-        for s in &mut self.syntaxes {
-            if s.glob_hash == syntax.glob_hash {
-                *s = syntax;
-                return;
+    pub fn set_current_from_glob(&mut self, glob: &str) -> Result<(), InvalidGlobError> {
+        let glob_hash = hash_bytes(glob.as_bytes());
+        for (i, s) in self.syntaxes.iter_mut().enumerate() {
+            if s.glob_hash == glob_hash {
+                s.clear_rules();
+                self.current_syntax_index = i as _;
+                return Ok(());
             }
         }
+
+        self.current_syntax_index = self.syntaxes.len() as _;
+        let mut syntax = Syntax::new();
+        syntax.set_glob(glob, glob_hash)?;
         self.syntaxes.push(syntax);
+        Ok(())
+    }
+
+    pub fn get_current(&mut self) -> &mut Syntax {
+        &mut self.syntaxes[self.current_syntax_index as usize]
     }
 
     pub fn get(&self, handle: SyntaxHandle) -> &Syntax {
-        &self.syntaxes[handle.0]
+        &self.syntaxes[handle.0 as usize]
     }
 }
 
@@ -258,15 +289,6 @@ pub struct HighlightedBuffer {
 }
 
 impl HighlightedBuffer {
-    pub fn empty() -> &'static Self {
-        static EMPTY: HighlightedBuffer = HighlightedBuffer {
-            highlighted_len: 0,
-            lines: Vec::new(),
-            dirty_line_indexes: Vec::new(),
-        };
-        &EMPTY
-    }
-
     pub fn new() -> Self {
         Self {
             highlighted_len: 1,
@@ -280,7 +302,7 @@ impl HighlightedBuffer {
         self.dirty_line_indexes.clear();
     }
 
-    pub fn on_insert(&mut self, range: BufferRange) {
+    pub fn insert_range(&mut self, range: BufferRange) {
         let insert_line_count = range.to.line_index - range.from.line_index;
         if insert_line_count > 0 {
             let previous_highlighted_len = self.highlighted_len;
@@ -312,7 +334,7 @@ impl HighlightedBuffer {
         self.dirty_line_indexes.push(range.from.line_index);
     }
 
-    pub fn on_delete(&mut self, range: BufferRange) {
+    pub fn delete_range(&mut self, range: BufferRange) {
         self.lines[range.from.line_index as usize].parse_state = LineParseState::Dirty;
 
         let delete_line_count = range.to.line_index - range.from.line_index;
@@ -342,9 +364,10 @@ impl HighlightedBuffer {
             return HighlightResult::Complete;
         }
 
-        self.dirty_line_indexes.sort();
+        self.dirty_line_indexes.sort_unstable();
+
         let mut index = self.dirty_line_indexes[0];
-        let mut last_dirty_index = usize::MAX;
+        let mut last_dirty_index = BufferPositionIndex::MAX;
 
         let mut previous_parse_state = match index.checked_sub(1) {
             Some(i) => self.lines[i as usize].parse_state,
@@ -357,12 +380,12 @@ impl HighlightedBuffer {
             let dirty_index = self.dirty_line_indexes[i];
             i += 1;
 
-            if dirty_index < index || dirty_index == last_dirty_index as _ {
+            if dirty_index < index || dirty_index == last_dirty_index {
                 continue;
             }
 
             index = dirty_index;
-            last_dirty_index = dirty_index as _;
+            last_dirty_index = dirty_index;
 
             while index < self.highlighted_len as _ {
                 let bline = buffer.line_at(index as _).as_str();
@@ -379,7 +402,7 @@ impl HighlightedBuffer {
                 if highlight_count == MAX_HIGHLIGHT_COUNT {
                     i -= 1;
                     self.dirty_line_indexes[i] = index;
-                    self.dirty_line_indexes.rotate_left(i);
+                    self.dirty_line_indexes.drain(..i);
                     return HighlightResult::Pending;
                 }
 
@@ -494,6 +517,21 @@ mod tests {
         assert_token("  ", TokenKind::Text, line, &tokens[5]);
     }
 
+    #[test]
+    fn beginning_anchor_syntax() {
+        let mut syntax = Syntax::new();
+        syntax.set_rule(TokenKind::Keyword, "^{%w}").unwrap();
+
+        let mut tokens = Vec::new();
+        let line = "first second";
+        let parse_state = syntax.parse_line(line, LineParseState::Finished, &mut tokens);
+
+        assert_eq!(LineParseState::Finished, parse_state);
+        assert_eq!(2, tokens.len());
+        assert_token("first", TokenKind::Keyword, line, &tokens[0]);
+        assert_token(" second", TokenKind::Text, line, &tokens[1]);
+    }
+
     #[test]
     fn multiline_syntax() {
         let mut syntax = Syntax::new();
@@ -538,7 +576,7 @@ mod tests {
         let mut highlighted = HighlightedBuffer::new();
 
         let range = buffer.insert_text(BufferPosition::zero(), "/*\n*/");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
         assert_eq!(buffer.line_count(), highlighted.lines.len());
 
@@ -550,7 +588,7 @@ mod tests {
         }
 
         let range = buffer.insert_text(BufferPosition::line_col(1, 0), "'");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
 
         {
@@ -570,7 +608,7 @@ mod tests {
         let mut highlighted = HighlightedBuffer::new();
 
         let range = buffer.insert_text(BufferPosition::zero(), "/*\n\n\n*/");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
         assert_eq!(buffer.line_count(), highlighted.lines.len());
 
@@ -591,7 +629,7 @@ mod tests {
         let mut highlighted = HighlightedBuffer::new();
 
         let range = buffer.insert_text(BufferPosition::zero(), "/*\n* /\n*/");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
 
         let range = BufferRange::between(
@@ -599,7 +637,7 @@ mod tests {
             BufferPosition::line_col(1, 2),
         );
         buffer.delete_range(range);
-        highlighted.on_delete(range);
+        highlighted.delete_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
 
         let mut parse_states = highlighted.lines[..highlighted.highlighted_len]
@@ -632,7 +670,7 @@ mod tests {
         let mut highlighted = HighlightedBuffer::new();
 
         let range = buffer.insert_text(BufferPosition::zero(), "/ *\na\n*/");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
 
         let range = BufferRange::between(
@@ -640,7 +678,7 @@ mod tests {
             BufferPosition::line_col(0, 2),
         );
         buffer.delete_range(range);
-        highlighted.on_delete(range);
+        highlighted.delete_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
 
         let mut tokens = highlighted_tokens(&highlighted);
@@ -659,7 +697,7 @@ mod tests {
         let mut highlighted = HighlightedBuffer::new();
 
         let range = buffer.insert_text(BufferPosition::zero(), "a\n/*\nb\nc*/");
-        highlighted.on_insert(range);
+        highlighted.insert_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
         assert_eq!(buffer.line_count(), highlighted.highlighted_len);
 
@@ -674,7 +712,7 @@ mod tests {
 
         let range = BufferRange::between(BufferPosition::zero(), BufferPosition::line_col(1, 1));
         buffer.delete_range(range);
-        highlighted.on_delete(range);
+        highlighted.delete_range(range);
         highlighted.highlight_dirty_lines(&syntax, &buffer);
         assert_eq!(buffer.line_count(), highlighted.highlighted_len);
 
