@@ -1,14 +1,13 @@
 use std::{io, iter};
 
 use crate::{
-    buffer::{Buffer, BufferContent},
     buffer_position::{BufferPosition, BufferRange},
     buffer_view::{BufferViewHandle, CursorMovementKind},
-    cursor::Cursor,
+    client::ClientManager,
     editor::Editor,
     editor_utils::MessageKind,
     mode::ModeKind,
-    syntax::{HighlightedBuffer, Token, TokenKind},
+    syntax::{Token, TokenKind},
     theme::Color,
 };
 
@@ -23,175 +22,163 @@ pub static END_TITLE_CODE: &[u8] = b"\x07";
 
 static TOO_LONG_PREFIX: &[u8] = b"...";
 
-#[inline]
 pub fn clear_line(buf: &mut Vec<u8>) {
     buf.extend_from_slice(b"\x1b[2K");
 }
 
-#[inline]
 pub fn clear_until_new_line(buf: &mut Vec<u8>) {
     buf.extend_from_slice(b"\x1b[0K");
 }
 
-#[inline]
 pub fn move_cursor_to(buf: &mut Vec<u8>, x: usize, y: usize) {
     use io::Write;
     let _ = write!(buf, "\x1b[{};{}H", x, y);
 }
 
-#[inline]
 pub fn move_cursor_to_next_line(buf: &mut Vec<u8>) {
     buf.extend_from_slice(b"\x1b[1E");
 }
 
-#[inline]
 pub fn move_cursor_up(buf: &mut Vec<u8>, count: usize) {
     use io::Write;
     let _ = write!(buf, "\x1b[{}A", count);
 }
 
-#[inline]
 pub fn set_background_color(buf: &mut Vec<u8>, color: Color) {
     use io::Write;
     let _ = write!(buf, "\x1b[48;2;{};{};{}m", color.0, color.1, color.2);
 }
 
-#[inline]
 pub fn set_foreground_color(buf: &mut Vec<u8>, color: Color) {
     use io::Write;
     let _ = write!(buf, "\x1b[38;2;{};{};{}m", color.0, color.1, color.2);
 }
 
-#[inline]
 pub fn set_underlined(buf: &mut Vec<u8>) {
     buf.extend_from_slice(b"\x1b[4m");
 }
 
-#[inline]
 pub fn set_not_underlined(buf: &mut Vec<u8>) {
     buf.extend_from_slice(b"\x1b[24m");
 }
 
+pub struct RenderContext<'a> {
+    pub editor: &'a Editor,
+    pub clients: &'a ClientManager,
+    pub viewport_size: (u16, u16),
+    pub scroll: (u32, u32),
+    pub draw_height: u16,
+    pub has_focus: bool,
+}
+
+fn draw_empty_view(ctx: &RenderContext, buf: &mut Vec<u8>) {
+    move_cursor_to(buf, 0, 0);
+    buf.extend_from_slice(RESET_STYLE_CODE);
+    set_background_color(buf, ctx.editor.theme.background);
+    set_foreground_color(buf, ctx.editor.theme.token_whitespace);
+
+    let message_lines = &[
+        concat!(env!("CARGO_PKG_NAME"), " editor"),
+        concat!("version ", env!("CARGO_PKG_VERSION")),
+        "",
+        "type `:open <file-name><enter>` to edit a file",
+        "or `:help<enter>` for help",
+    ];
+
+    let width = ctx.viewport_size.0 as usize - 1;
+    let height = ctx.viewport_size.1 as usize - 1;
+    let draw_height = ctx.draw_height as usize;
+
+    let margin_top = (height.saturating_sub(message_lines.len())) / 2;
+    let margin_bottom = draw_height - margin_top - message_lines.len();
+
+    let mut visual_empty = [0; 4];
+    let visual_empty = ctx
+        .editor
+        .config
+        .visual_empty
+        .encode_utf8(&mut visual_empty)
+        .as_bytes();
+
+    for _ in 0..margin_top {
+        buf.extend_from_slice(visual_empty);
+        clear_until_new_line(buf);
+        move_cursor_to_next_line(buf);
+    }
+
+    for line in message_lines {
+        buf.extend_from_slice(visual_empty);
+
+        let margin_left = (width.saturating_sub(line.len())) / 2;
+        buf.extend(std::iter::repeat(b' ').take(margin_left));
+        buf.extend_from_slice(line.as_bytes());
+
+        clear_until_new_line(buf);
+        move_cursor_to_next_line(buf);
+    }
+
+    for _ in 0..margin_bottom {
+        buf.extend_from_slice(visual_empty);
+        clear_until_new_line(buf);
+        move_cursor_to_next_line(buf);
+    }
+}
+
 pub fn render(
-    editor: &Editor,
+    ctx: &RenderContext,
     buffer_view_handle: Option<BufferViewHandle>,
-    size: (u16, u16),
-    scroll: (u32, u32),
-    has_focus: bool,
-    buffer: &mut Vec<u8>,
+    buf: &mut Vec<u8>,
 ) {
-    let view = View::new(editor, buffer_view_handle, size, scroll);
-
-    draw_buffer(buffer, editor, &view, has_focus);
-    if has_focus {
-        draw_picker(buffer, editor, &view);
-    }
-    draw_statusbar(buffer, editor, &view, has_focus);
+    draw_buffer_view(ctx, buffer_view_handle, buf);
+    draw_picker(ctx, buf);
+    draw_statusbar(ctx, buffer_view_handle, buf);
 }
 
-struct View<'a> {
-    buffer: Option<&'a Buffer>,
-    main_cursor_position: BufferPosition,
-    cursors: &'a [Cursor],
-
-    size: (u16, u16),
-    scroll: (u32, u32),
-}
-
-impl<'a> View<'a> {
-    pub fn new(
-        editor: &'a Editor,
-        buffer_view_handle: Option<BufferViewHandle>,
-        size: (u16, u16),
-        scroll: (u32, u32),
-    ) -> View<'a> {
-        let buffer_view = buffer_view_handle.and_then(|h| editor.buffer_views.get(h));
-        let buffer_handle = buffer_view.map(|v| v.buffer_handle);
-        let buffer = buffer_handle.and_then(|h| editor.buffers.get(h));
-
-        let main_cursor_position;
-        let cursors;
-        match buffer_view {
-            Some(view) => {
-                main_cursor_position = view.cursors.main_cursor().position;
-                cursors = &view.cursors[..];
-            }
-            None => {
-                main_cursor_position = BufferPosition::default();
-                cursors = &[];
-            }
-        };
-
-        View {
-            buffer,
-            main_cursor_position,
-            cursors,
-            size,
-            scroll,
+fn draw_buffer_view(
+    ctx: &RenderContext,
+    buffer_view_handle: Option<BufferViewHandle>,
+    buf: &mut Vec<u8>,
+) {
+    let buffer_view_handle = match buffer_view_handle {
+        Some(handle) => handle,
+        None => {
+            draw_empty_view(ctx, buf);
+            return;
         }
-    }
-}
+    };
 
-fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool) {
-    let mut char_buf = [0; std::mem::size_of::<char>()];
+    let buffer_view = ctx.editor.buffer_views.get(buffer_view_handle);
+    let buffer = ctx.editor.buffers.get(buffer_view.buffer_handle);
+    let cursors = &buffer_view.cursors[..];
+    let active_line_index = buffer_view.cursors.main_cursor().position.line_index as usize;
 
-    let cursor_color = if has_focus {
-        match editor.mode.kind() {
-            ModeKind::Insert => editor.theme.insert_cursor,
-            _ => match editor.mode.normal_state.movement_kind {
-                CursorMovementKind::PositionAndAnchor => editor.theme.normal_cursor,
-                CursorMovementKind::PositionOnly => editor.theme.select_cursor,
+    let cursor_color = if ctx.has_focus {
+        match ctx.editor.mode.kind() {
+            ModeKind::Insert => ctx.editor.theme.insert_cursor,
+            _ => match ctx.editor.mode.normal_state.movement_kind {
+                CursorMovementKind::PositionAndAnchor => ctx.editor.theme.normal_cursor,
+                CursorMovementKind::PositionOnly => ctx.editor.theme.select_cursor,
             },
         }
     } else {
-        editor.theme.inactive_cursor
+        ctx.editor.theme.inactive_cursor
     };
 
-    move_cursor_to(buf, 0, 0);
-    set_background_color(buf, editor.theme.background);
-    set_not_underlined(buf);
-
-    let cursors = &view.cursors[..];
     let cursors_end_index = cursors.len().saturating_sub(1);
-    let active_line_index = view.main_cursor_position.line_index;
 
-    let buffer_content;
-    let highlighted_buffer;
-    let search_ranges;
-    match view.buffer {
-        Some(buffer) => {
-            buffer_content = buffer.content();
-            highlighted_buffer = buffer.highlighted();
-            search_ranges = buffer.search_ranges();
-        }
-        None => {
-            buffer_content = BufferContent::empty();
-            highlighted_buffer = HighlightedBuffer::empty();
-            search_ranges = &[];
-        }
-    }
+    let buffer_content = buffer.content();
+    let highlighted_buffer = buffer.highlighted();
+    let search_ranges = buffer.search_ranges();
     let search_ranges_end_index = search_ranges.len().saturating_sub(1);
 
-    let diagnostics = match view.buffer {
-        Some(buffer) => {
-            let mut diagnostics: &[_] = &[];
-            for client in editor.lsp.clients() {
-                diagnostics = client.diagnostics().buffer_diagnostics(buffer.handle());
-                if !diagnostics.is_empty() {
-                    break;
-                }
-            }
-            diagnostics
-        }
-        None => &[],
-    };
-    let diagnostics_end_index = diagnostics.len().saturating_sub(1);
+    let lints = buffer.lints.all();
+    let lints_end_index = lints.len().saturating_sub(1);
 
-    let display_position_offset = BufferPosition::line_col(view.scroll.1 as _, view.scroll.0 as _);
+    let display_position_offset = BufferPosition::line_col(ctx.scroll.1 as _, ctx.scroll.0 as _);
 
     let mut current_cursor_index = cursors.len();
-    let mut current_cursor_position = BufferPosition::default();
-    let mut current_cursor_range = BufferRange::default();
+    let mut current_cursor_position = BufferPosition::zero();
+    let mut current_cursor_range = BufferRange::zero();
     for (i, cursor) in cursors.iter().enumerate() {
         let range = cursor.to_range();
         if display_position_offset <= range.to {
@@ -203,7 +190,7 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
     }
 
     let mut current_search_range_index = search_ranges.len();
-    let mut current_search_range = BufferRange::default();
+    let mut current_search_range = BufferRange::zero();
     for (i, &range) in search_ranges.iter().enumerate() {
         if display_position_offset < range.to {
             current_search_range_index = i;
@@ -212,22 +199,60 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
         }
     }
 
-    let mut current_diagnostic_index = diagnostics.len();
-    let mut current_diagnostic_range = BufferRange::default();
-    for (i, diagnostic) in diagnostics.iter().enumerate() {
-        if display_position_offset < diagnostic.range.to {
-            current_diagnostic_index = i;
-            current_diagnostic_range = diagnostic.range;
+    let mut current_lint_index = lints.len();
+    let mut current_lint_range = BufferRange::zero();
+    for (i, lint) in lints.iter().enumerate() {
+        if display_position_offset < lint.range.to {
+            current_lint_index = i;
+            current_lint_range = lint.range;
             break;
         }
     }
 
+    move_cursor_to(buf, 0, 0);
+    set_background_color(buf, ctx.editor.theme.background);
+    set_not_underlined(buf);
+
+    let mut char_buf = [0; std::mem::size_of::<char>()];
+
+    let mut visual_empty = [0; 4];
+    let visual_empty = ctx
+        .editor
+        .config
+        .visual_empty
+        .encode_utf8(&mut visual_empty)
+        .as_bytes();
+
+    let mut visual_space = [0; 4];
+    let visual_space = ctx
+        .editor
+        .config
+        .visual_space
+        .encode_utf8(&mut visual_space)
+        .as_bytes();
+
+    let mut visual_tab_first = [0; 4];
+    let visual_tab_first = ctx
+        .editor
+        .config
+        .visual_tab_first
+        .encode_utf8(&mut visual_tab_first)
+        .as_bytes();
+
+    let mut visual_tab_repeat = [0; 4];
+    let visual_tab_repeat = ctx
+        .editor
+        .config
+        .visual_tab_repeat
+        .encode_utf8(&mut visual_tab_repeat)
+        .as_bytes();
+
     let mut lines_drawn_count = 0;
     for (line_index, line) in buffer_content
         .lines()
         .enumerate()
-        .skip(view.scroll.1 as _)
-        .take(view.size.1 as _)
+        .skip(ctx.scroll.1 as _)
+        .take(ctx.draw_height as _)
     {
         #[derive(Clone, Copy, PartialEq, Eq)]
         enum DrawState {
@@ -241,22 +266,22 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
 
         let line = line.as_str();
         let mut draw_state = DrawState::Token(TokenKind::Text);
-        let mut was_inside_diagnostic_range = false;
+        let mut was_inside_lint_range = false;
         let mut x = 0;
         let mut last_line_token = Token::default();
         let mut line_tokens = highlighted_buffer.line_tokens(line_index).iter();
 
         let background_color = if line_index == active_line_index as _ {
-            editor.theme.active_line_background
+            ctx.editor.theme.active_line_background
         } else {
-            editor.theme.background
+            ctx.editor.theme.background
         };
 
         set_background_color(buf, background_color);
-        set_foreground_color(buf, editor.theme.token_text);
+        set_foreground_color(buf, ctx.editor.theme.token_text);
 
         for (char_index, c) in line.char_indices().chain(iter::once((line.len(), '\n'))) {
-            if char_index < view.scroll.0 as _ {
+            if char_index < ctx.scroll.0 as _ {
                 continue;
             }
 
@@ -278,14 +303,14 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
             };
 
             let text_color = match token_kind {
-                TokenKind::Keyword => editor.theme.token_keyword,
-                TokenKind::Type => editor.theme.token_type,
-                TokenKind::Symbol => editor.theme.token_symbol,
-                TokenKind::Literal => editor.theme.token_literal,
-                TokenKind::String => editor.theme.token_string,
-                TokenKind::Comment => editor.theme.token_comment,
-                TokenKind::Text => editor.theme.token_text,
-                TokenKind::Whitespace => editor.theme.token_whitespace,
+                TokenKind::Keyword => ctx.editor.theme.token_keyword,
+                TokenKind::Type => ctx.editor.theme.token_type,
+                TokenKind::Symbol => ctx.editor.theme.token_symbol,
+                TokenKind::Literal => ctx.editor.theme.token_literal,
+                TokenKind::String => ctx.editor.theme.token_string,
+                TokenKind::Comment => ctx.editor.theme.token_comment,
+                TokenKind::Text => ctx.editor.theme.token_text,
+                TokenKind::Whitespace => ctx.editor.theme.token_whitespace,
             };
 
             while current_cursor_index < cursors_end_index
@@ -308,18 +333,16 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
             let inside_search_range = current_search_range.from <= char_position
                 && char_position < current_search_range.to;
 
-            while current_diagnostic_range.to < char_position
-                && current_diagnostic_index < diagnostics_end_index
-            {
-                current_diagnostic_index += 1;
-                current_diagnostic_range = diagnostics[current_diagnostic_index].range;
+            while current_lint_range.to < char_position && current_lint_index < lints_end_index {
+                current_lint_index += 1;
+                current_lint_range = lints[current_lint_index].range;
             }
-            let inside_diagnostic_range = current_diagnostic_range.from <= char_position
-                && char_position < current_diagnostic_range.to;
+            let inside_lint_range =
+                current_lint_range.from <= char_position && char_position < current_lint_range.to;
 
-            if inside_diagnostic_range != was_inside_diagnostic_range {
-                was_inside_diagnostic_range = inside_diagnostic_range;
-                if inside_diagnostic_range {
+            if inside_lint_range != was_inside_lint_range {
+                was_inside_lint_range = inside_lint_range;
+                if inside_lint_range {
                     set_underlined(buf);
                 } else {
                     set_not_underlined(buf);
@@ -341,7 +364,7 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
             } else if inside_search_range {
                 if draw_state != DrawState::Highlight {
                     draw_state = DrawState::Highlight;
-                    set_background_color(buf, editor.theme.highlight);
+                    set_background_color(buf, ctx.editor.theme.highlight);
                     set_foreground_color(buf, background_color);
                 }
             } else if draw_state != DrawState::Token(token_kind) {
@@ -358,15 +381,15 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
                 }
                 ' ' => {
                     x += 1;
-                    buf.push(editor.config.visual_space);
+                    buf.extend_from_slice(visual_space);
                 }
                 '\t' => {
-                    let tab_size = editor.config.tab_size.get() as usize;
+                    let tab_size = ctx.editor.config.tab_size.get() as usize;
                     x += tab_size;
 
-                    buf.push(editor.config.visual_tab_first);
+                    buf.extend_from_slice(visual_tab_first);
                     for _ in 0..tab_size - 1 {
-                        buf.push(editor.config.visual_tab_repeat);
+                        buf.extend_from_slice(visual_tab_repeat);
                     }
                 }
                 _ => {
@@ -375,7 +398,7 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
                 }
             }
 
-            if x > view.size.0 as _ {
+            if x > ctx.viewport_size.0 as _ {
                 x = previous_x;
                 buf.truncate(buf_len);
                 break;
@@ -384,7 +407,7 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
 
         set_background_color(buf, background_color);
 
-        if x < view.size.0 as _ {
+        if x < ctx.viewport_size.0 as _ {
             clear_until_new_line(buf);
         }
 
@@ -392,36 +415,42 @@ fn draw_buffer(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool)
     }
 
     set_not_underlined(buf);
-    set_background_color(buf, editor.theme.background);
-    set_foreground_color(buf, editor.theme.token_whitespace);
+    set_background_color(buf, ctx.editor.theme.background);
+    set_foreground_color(buf, ctx.editor.theme.token_whitespace);
 
-    for _ in lines_drawn_count..view.size.1 {
-        buf.push(editor.config.visual_empty);
+    for _ in lines_drawn_count..ctx.draw_height {
+        buf.extend_from_slice(visual_empty);
         clear_until_new_line(buf);
         move_cursor_to_next_line(buf);
     }
 }
 
-fn draw_picker(buf: &mut Vec<u8>, editor: &Editor, view: &View) {
-    let cursor = editor.picker.cursor().unwrap_or(usize::MAX - 1);
-    let scroll = editor.picker.scroll();
+fn draw_picker(ctx: &RenderContext, buf: &mut Vec<u8>) {
+    if !ctx.has_focus {
+        return;
+    }
 
-    let width = view.size.0 as _;
-    let height = editor
+    let cursor = ctx.editor.picker.cursor().unwrap_or(usize::MAX - 1);
+    let scroll = ctx.editor.picker.scroll();
+
+    let width = ctx.viewport_size.0 as _;
+    let height = ctx
+        .editor
         .picker
         .len()
-        .min(editor.config.picker_max_height as _);
+        .min(ctx.editor.config.picker_max_height as _);
 
-    let background_normal_color = editor.theme.statusbar_inactive_background;
-    let background_selected_color = editor.theme.statusbar_active_background;
-    let foreground_color = editor.theme.token_text;
+    let background_normal_color = ctx.editor.theme.statusbar_inactive_background;
+    let background_selected_color = ctx.editor.theme.statusbar_active_background;
+    let foreground_color = ctx.editor.theme.token_text;
 
     set_background_color(buf, background_normal_color);
     set_foreground_color(buf, foreground_color);
 
-    for (i, entry) in editor
+    for (i, entry) in ctx
+        .editor
         .picker
-        .entries(&editor.word_database)
+        .entries(&ctx.editor.word_database)
         .enumerate()
         .skip(scroll)
         .take(height)
@@ -434,7 +463,6 @@ fn draw_picker(buf: &mut Vec<u8>, editor: &Editor, view: &View) {
 
         let mut x = 0;
 
-        #[inline]
         fn print_char(buf: &mut Vec<u8>, x: &mut usize, c: char) {
             let mut char_buf = [0; std::mem::size_of::<char>()];
 
@@ -470,28 +498,55 @@ fn draw_picker(buf: &mut Vec<u8>, editor: &Editor, view: &View) {
     }
 }
 
-fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bool) {
+fn draw_statusbar(
+    ctx: &RenderContext,
+    buffer_view_handle: Option<BufferViewHandle>,
+    buf: &mut Vec<u8>,
+) {
+    let view_name;
+    let needs_save;
+    let main_cursor_position;
+    let search_ranges;
+
+    match buffer_view_handle {
+        Some(handle) => {
+            let buffer_view = ctx.editor.buffer_views.get(handle);
+            let buffer = ctx.editor.buffers.get(buffer_view.buffer_handle);
+
+            view_name = buffer.path.to_str().unwrap_or("");
+            needs_save = buffer.needs_save();
+            main_cursor_position = buffer_view.cursors.main_cursor().position;
+            search_ranges = buffer.search_ranges();
+        }
+        None => {
+            view_name = "";
+            needs_save = false;
+            main_cursor_position = BufferPosition::zero();
+            search_ranges = &[];
+        }
+    }
+
     use io::Write;
 
-    let background_active_color = editor.theme.statusbar_active_background;
-    let background_innactive_color = editor.theme.statusbar_inactive_background;
-    let foreground_color = editor.theme.token_text;
-    let cursor_color = editor.theme.normal_cursor;
+    let background_active_color = ctx.editor.theme.statusbar_active_background;
+    let background_innactive_color = ctx.editor.theme.statusbar_inactive_background;
+    let foreground_color = ctx.editor.theme.token_text;
+    let cursor_color = ctx.editor.theme.normal_cursor;
 
-    if has_focus {
+    if ctx.has_focus {
         set_background_color(buf, background_active_color);
     } else {
         set_background_color(buf, background_innactive_color);
     }
     set_foreground_color(buf, foreground_color);
 
-    let x = if has_focus {
-        let (message_target, message) = editor.status_bar.message();
+    let x = if ctx.has_focus {
+        let (message_target, message) = ctx.editor.status_bar.message();
         let message = message.trim_end();
 
         let message_is_empty = message.trim_start().is_empty();
-        match editor.mode.kind() {
-            ModeKind::Normal if message_is_empty => match editor.recording_macro {
+        match ctx.editor.mode.kind() {
+            ModeKind::Normal if message_is_empty => match ctx.editor.recording_macro {
                 Some(key) => {
                     let text = b"recording macro ";
                     let key = key.as_u8();
@@ -499,11 +554,11 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
                     buf.push(key);
                     Some(text.len() + 1)
                 }
-                None => match view.buffer.map(Buffer::search_ranges) {
-                    Some(&[]) | None => Some(0),
-                    Some(search_ranges) => {
+                None => match search_ranges {
+                    [] => Some(0),
+                    _ => {
                         let previous_len = buf.len();
-                        let search_index = editor.mode.normal_state.search_index + 1;
+                        let search_index = ctx.editor.mode.normal_state.search_index + 1;
                         let _ = write!(buf, " [{}/{}]", search_index, search_ranges.len());
                         Some(buf.len() - previous_len)
                     }
@@ -515,7 +570,7 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
                 Some(text.len())
             }
             ModeKind::Command | ModeKind::Picker | ModeKind::ReadLine => {
-                let read_line = &editor.read_line;
+                let read_line = &ctx.editor.read_line;
 
                 set_background_color(buf, background_innactive_color);
                 set_foreground_color(buf, foreground_color);
@@ -534,15 +589,10 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
                     let mut len = 0;
                     for c in line.chars() {
                         match c {
-                            '\t' => {
-                                buf.extend_from_slice(b"  ");
-                                len += 2;
-                            }
-                            c => {
-                                buf.extend_from_slice(c.encode_utf8(&mut char_buf).as_bytes());
-                                len += 1;
-                            }
-                        };
+                            '\t' => buf.push(b' '),
+                            c => buf.extend_from_slice(c.encode_utf8(&mut char_buf).as_bytes()),
+                        }
+                        len += 1;
                     }
                     len
                 }
@@ -552,7 +602,24 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
                     MessageKind::Error => &b"error:"[..],
                 };
 
-                let line_count = message.lines().count();
+                let line_count = {
+                    let mut x = 0;
+                    let mut line_count = 1;
+                    for c in message.chars() {
+                        if c == '\n' {
+                            x = 0;
+                            line_count += 1;
+                        } else {
+                            x += 1;
+                            if x >= ctx.viewport_size.0 as _ {
+                                x = 0;
+                                line_count += 1;
+                            }
+                        }
+                    }
+                    line_count
+                };
+
                 if line_count > 1 {
                     if prefix.is_empty() {
                         move_cursor_up(buf, line_count - 1);
@@ -570,7 +637,7 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
                     for (i, line) in message.lines().enumerate() {
                         let len = print_line(buf, line);
                         if i < line_count - 1 {
-                            if len < view.size.0 as _ {
+                            if len == 0 || len % ctx.viewport_size.0 as usize > 0 {
                                 clear_until_new_line(buf);
                             }
                             move_cursor_to_next_line(buf);
@@ -593,19 +660,6 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
         Some(0)
     };
 
-    let buffer_needs_save;
-    let buffer_path;
-    match view.buffer {
-        Some(buffer) => {
-            buffer_needs_save = buffer.needs_save();
-            buffer_path = buffer.path.to_str().unwrap_or("<no path>");
-        }
-        None => {
-            buffer_needs_save = false;
-            buffer_path = "<no buffer>";
-        }
-    };
-
     if let Some(x) = x {
         fn take_chars(s: &str, char_count: usize) -> (usize, &str) {
             match s.char_indices().rev().enumerate().take(char_count).last() {
@@ -614,35 +668,35 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
             }
         }
 
-        let available_width = view.size.0 as usize - x;
+        let available_width = ctx.viewport_size.0 as usize - x;
         let half_available_width = available_width / 2;
 
         let status_start_index = buf.len();
 
-        if has_focus {
-            let param_count = editor.mode.normal_state.count;
-            if param_count > 0 && matches!(editor.mode.kind(), ModeKind::Normal) {
+        if ctx.has_focus {
+            let param_count = ctx.editor.mode.normal_state.count;
+            if param_count > 0 && matches!(ctx.editor.mode.kind(), ModeKind::Normal) {
                 let _ = write!(buf, "{}", param_count);
             }
-            for key in editor.buffered_keys.as_slice() {
+            for key in ctx.editor.buffered_keys.as_slice() {
                 let _ = write!(buf, "{}", key);
             }
             buf.push(b' ');
         }
 
-        if buffer_needs_save {
+        if needs_save {
             buf.push(b'*');
         }
 
-        let (char_count, buffer_path) = take_chars(buffer_path, half_available_width);
+        let (char_count, view_name) = take_chars(view_name, half_available_width);
         if char_count == half_available_width {
             buf.extend_from_slice(TOO_LONG_PREFIX);
         }
-        buf.extend_from_slice(buffer_path.as_bytes());
+        buf.extend_from_slice(view_name.as_bytes());
 
-        if view.buffer.is_some() {
-            let line_number = view.main_cursor_position.line_index + 1;
-            let column_number = view.main_cursor_position.column_byte_index + 1;
+        if !view_name.is_empty() {
+            let line_number = main_cursor_position.line_index + 1;
+            let column_number = main_cursor_position.column_byte_index + 1;
             let _ = write!(buf, ":{},{}", line_number, column_number);
         }
         buf.push(b' ');
@@ -679,10 +733,14 @@ fn draw_statusbar(buf: &mut Vec<u8>, editor: &Editor, view: &View, has_focus: bo
     }
 
     buf.extend_from_slice(BEGIN_TITLE_CODE);
-    if buffer_needs_save {
+    if needs_save {
         buf.push(b'*');
     }
-    buf.extend_from_slice(buffer_path.as_bytes());
+    if view_name.is_empty() {
+        buf.extend_from_slice(b"no buffer");
+    } else {
+        buf.extend_from_slice(view_name.as_bytes());
+    }
     buf.extend_from_slice(END_TITLE_CODE);
 
     clear_until_new_line(buf);
